'+--------------------------------------------------------------------------+
'|                                                                          |
'|                                  BUFIN                                   |
'|                                                                          |
'|    Fast line input replacement based on the core Windows API file I/O    |
'|                                                                          |
'|               Adapted from the WinLIFT's skBufin function                |                                                                          |
'|                                                                          |
'+--------------------------------------------------------------------------+
'|                                                                          |
'|                         Author Patrice TERRIER                           |
'|            8 Domaine de Rochagnon. 38800 Champagnier  FRANCE             |
'|                          www.zapsolution.com                             |
'|                       pterrier@zapsolution.com                           |
'|                                                                          |
'|                copyright (c) 2000-2007 Patrice TERRIER                   |
'|                                                                          |
'+--------------------------------------------------------------------------+
'|                  Project started on : 04-19-2000                         |
'|                        Last revised : 04-19-2000                         |
'+--------------------------------------------------------------------------+
'
' This example shows how to use the Bufin$ function to read
' in a sequential ASCII text file.
' Text line can be delimited with a single CR or both CR+LF.
'
' This example works in console mode, you must use the MKCON.EXE utility
' to make it work from a DOS windows session.

#COMPILE EXE

'-----------------------------------------------------------------
' Equates:  14
'-----------------------------------------------------------------
%WINAPI                                         = 1
%NULL                                           = 0
%GENERIC_READ                                   = &H80000000&
%GENERIC_WRITE                                  = &H40000000&
%FILE_SHARE_READ                                = &H00000001
%FILE_SHARE_WRITE                               = &H00000002
%FILE_ATTRIBUTE_NORMAL                          = &H00000080
%INVALID_HANDLE_VALUE                           = &HFFFFFFFF???
%FILE_BEGIN                                     = 0
%FILE_FLAG_WRITE_THROUGH                        = &H080000000
%OPEN_ALWAYS                                    = 4
%FILE_TYPE_DISK                                 = &H1
%STD_OUTPUT_HANDLE                              = -11&
%MAX_PATH                                       = 260  ' max. length of full pathname

'-----------------------------------------------------------------
' TYPE and UNION structures:  4
'-----------------------------------------------------------------
TYPE OVERLAPPED
  Internal AS DWORD
  InternalHigh AS DWORD
  offset AS DWORD
  OffsetHigh AS DWORD
  hEvent AS DWORD
END TYPE

TYPE SECURITY_ATTRIBUTES
  nLength AS DWORD
  lpSecurityDescriptor AS LONG
  bInheritHandle AS LONG
END TYPE

TYPE FILETIME
  dwLowDateTime AS DWORD
  dwHighDateTime AS DWORD
END TYPE

TYPE WIN32_FIND_DATA
  dwFileAttributes AS DWORD
  ftCreationTime AS FILETIME
  ftLastAccessTime AS FILETIME
  ftLastWriteTime AS FILETIME
  nFileSizeHigh AS DWORD
  nFileSizeLow AS DWORD
  dwReserved0 AS DWORD
  dwReserved1 AS DWORD
  cFileName AS ASCIIZ * %MAX_PATH
  cAlternateFileName AS ASCIIZ * 14
END TYPE

'-----------------------------------------------------------------
' Declared Functions:  11
'-----------------------------------------------------------------
DECLARE FUNCTION CloseHandle LIB "KERNEL32.DLL" ALIAS "CloseHandle" (BYVAL hObject AS DWORD) AS LONG
DECLARE FUNCTION CreateFile LIB "KERNEL32.DLL" ALIAS "CreateFileA" (lpFileName AS ASCIIZ, BYVAL dwDesiredAccess AS DWORD, BYVAL dwShareMode AS DWORD, lpSecurityAttributes AS SECURITY_ATTRIBUTES, BYVAL dwCreationDisposition AS DWORD, _
                 BYVAL dwFlagsAndAttributes AS DWORD, BYVAL hTemplateFile AS DWORD) AS DWORD
DECLARE FUNCTION FindClose LIB "KERNEL32.DLL" ALIAS "FindClose" (BYVAL hFindFile AS DWORD) AS LONG
DECLARE FUNCTION FindFirstFile LIB "KERNEL32.DLL" ALIAS "FindFirstFileA" (lpFileName AS ASCIIZ, lpFindFileData AS WIN32_FIND_DATA) AS DWORD
DECLARE FUNCTION GetFileSize LIB "KERNEL32.DLL" ALIAS "GetFileSize" (BYVAL hFile AS DWORD, lpFileSizeHigh AS LONG) AS LONG
DECLARE FUNCTION GetFileType LIB "KERNEL32.DLL" ALIAS "GetFileType" (BYVAL hFile AS DWORD) AS LONG
DECLARE FUNCTION GetLastError LIB "KERNEL32.DLL" ALIAS "GetLastError" () AS LONG
DECLARE FUNCTION GetStdHandle LIB "KERNEL32.DLL" ALIAS "GetStdHandle" (BYVAL nStdHandle AS DWORD) AS DWORD
DECLARE FUNCTION ReadFile LIB "KERNEL32.DLL" ALIAS "ReadFile" (BYVAL hFile AS DWORD, lpBuffer AS ANY, BYVAL nNumberOfBytesToRead AS DWORD, lpNumberOfBytesRead AS DWORD, lpOverlapped AS OVERLAPPED) AS LONG
DECLARE FUNCTION SetFilePointer LIB "KERNEL32.DLL" ALIAS "SetFilePointer" (BYVAL hFile AS DWORD, BYVAL lDistanceToMove AS LONG, lpDistanceToMoveHigh AS LONG, BYVAL dwMoveMethod AS DWORD) AS LONG
DECLARE FUNCTION WriteConsole LIB "KERNEL32.DLL" ALIAS "WriteConsoleA" (BYVAL hConsoleOutput AS DWORD, lpBuffer AS ASCIIZ, BYVAL nNumberOfCharsToWrite AS LONG, lpNumberOfCharsWritten AS LONG, BYVAL lpReserved AS LONG) AS LONG
DECLARE FUNCTION GetStdHandle LIB "KERNEL32.DLL" ALIAS "GetStdHandle" (BYVAL nStdHandle AS LONG) AS LONG
DECLARE FUNCTION WriteConsole LIB "KERNEL32.DLL" ALIAS "WriteConsoleA" (BYVAL hConsoleOutput AS LONG, lpBuffer AS ASCIIZ, BYVAL nNumberOfCharsToWrite AS LONG, lpNumberOfCharsWritten AS LONG, BYVAL lpReserved AS LONG) AS LONG

%STD_OUTPUT_HANDLE    = -11&

' Determines the presence of a file.
FUNCTION zExist ALIAS "zExist" (zFileSpec AS ASCIIZ) EXPORT AS LONG
    LOCAL fd AS WIN32_FIND_DATA
    LOCAL hFind AS LONG
    IF LEN(zFileSpec) THEN
       hFind& = FindFirstFile(zFileSpec, fd)
       IF hFind& <> %INVALID_HANDLE_VALUE THEN
          CALL FindClose(hFind)
          FUNCTION = -1
       END IF
    END IF
END FUNCTION

' Changes the character being used as delimiter to read in a sequential text file.
FUNCTION zSetBufinChar ALIAS "zSetBufinChar" (BYVAL Char$) STATIC EXPORT AS STRING
    IF LEN(Char$) THEN UserChar$ = Char$
    IF LEN(UserChar$) THEN
       FUNCTION = UserChar$
    ELSE
       FUNCTION = CHR$(13)
    END IF
END FUNCTION

' Generic open file function
FUNCTION zFOpen ALIAS "zFOpen"(zFilName AS ASCIIZ, BYVAL AccessMode AS LONG, BYVAL ShareMode AS LONG, hFile AS LONG) EXPORT AS LONG

    LOCAl AccessIs, ShareIs, FlagAndAttribute AS LONG

    AccessMode = MIN(MAX(AccessMode, 0), 2)  ' Coherce between 0-2
    IF AccessMode = 0 THEN                   ' 0 Open for read only.
       AccessIs = %GENERIC_READ
    ELSEIF AccessMode = 1 THEN               ' 1 Open for write only.
       AccessIs = %GENERIC_WRITE
    ELSE                                     ' 2 Open for read and write.
       AccessIs = %GENERIC_READ OR %GENERIC_WRITE
    END IF

    ShareMode = MIN(MAX(ShareMode, 1), 4)    ' Coherce between 1-4
    IF ShareMode = 1 THEN                    ' 1 Deny read/write access.
       ShareIs = 0
    ELSEIF ShareMode = 2 THEN                ' 2 Deny write access.
       ShareIs =  %FILE_SHARE_READ
    ELSEIF ShareMode = 3 THEN                ' 3 Deny read access.
       ShareIs =  %FILE_SHARE_WRITE
    ELSE                                     ' 4 Deny none (full share mode).
       ShareIs =  %FILE_SHARE_READ OR %FILE_SHARE_WRITE
    END IF

    IF hFile = -1 THEN
       FlagAndAttribute = %FILE_ATTRIBUTE_NORMAL OR %FILE_FLAG_WRITE_THROUGH
    ELSE
       FlagAndAttribute = %FILE_ATTRIBUTE_NORMAL
    END IF

    hFile = CreateFile(zFilName, AccessIs, ShareIs, BYVAL %NULL, %OPEN_ALWAYS, FlagAndAttribute, BYVAL %NULL)

    IF hFile = %INVALID_HANDLE_VALUE THEN    ' -1 Fail to create the file
       FUNCTION = GetLastError               ' Set the error code
       hFile = 0                             ' Reset handle number
    END IF

END FUNCTION

' Dito LOF
FUNCTION zFlof ALIAS "zFlof" (BYVAL hFile AS LONG) EXPORT AS LONG
    IF GetFileType(hFile) = %FILE_TYPE_DISK THEN
       LOCAL fSize AS LONG
       fSize = GetFileSize(hFile, BYVAL %NULL)
       IF fSize > -1 THEN FUNCTION = fSize
    END IF
END FUNCTION

' Dito SEEK
FUNCTION zFSeek ALIAS "zFSeek" (BYVAL hFile AS LONG, BYVAL PosByte AS LONG) EXPORT AS LONG
    IF SetFilePointer(hFile, PosByte, BYVAL %NULL, %FILE_BEGIN) < 0 THEN
       FUNCTION = GetLastError
    END IF
END FUNCTION

' Dito GET
FUNCTION zFGet ALIAS "zFGet" (BYVAL hFile AS LONG, StringBuf AS STRING) EXPORT AS LONG
    IF hFile THEN
       LOCAL LenBuf, ByttesReaded AS LONG
       LenBuf = LEN(StringBuf)
       IF LenBuf THEN
          IF ReadFile(hFile, BYVAL STRPTR(StringBuf), LenBuf, ByttesReaded, BYVAL %NULL) = 0 THEN
             FUNCTION = GetLastError
          END IF
       END IF
    END IF
END FUNCTION

' GET at a specific offset
FUNCTION zFGetAt ALIAS "zFGetAt" (BYVAL hFile AS LONG, BYVAL nPosByte AS LONG, Buf$) EXPORT AS LONG
    LOCAL nErrCode AS LONG
    nErrCode = zFSeek(hFile, nPosByte)
    IF nErrCode = 0 THEN nErrCode = zFGet(hFile, Buf$)
    FUNCTION = nErrCode
END FUNCTION

' Dito CLOSE
SUB zFClose ALIAS "zFClose" (hFile AS LONG) EXPORT
    IF hFile THEN CALL CloseHandle(hFile)
    hFile = 0
END SUB

' Enhanced sequential LINEINPUT
FUNCTION zBufin ALIAS "zBufin" (zFileName AS ASCIIZ, nDone AS LONG) STATIC EXPORT AS STRING
DIM sCr AS STRING, sBuffer AS STRING
DIM hFile AS LONG, nLocation AS LONG, nBufSize AS LONG, nBufPos AS LONG, nErrCode AS LONG, nRemaining AS LONG
DIM nCr AS LONG, nSlop AS LONG, nReading AS LONG, nLenCr AS LONG, nSizeFile AS LONG, nSaveCr AS LONG
IF LEN(zFileName) THEN
   IF NOT nReading THEN
      sCr = zSetBufinChar("")           ' added for spec char
      nReading = -1: nDone = 0
      IF ASC(sCr) = 13 THEN             ' added for spec char
         nLenCr = 0: sCr = CHR$(13)
      ELSE                              ' added for spec char
         nLenCr = LEN(sCr)              ' added for spec char
      END IF                            ' added for spec char
      nCr = 0: nLocation = 0
      IF NOT zExist(zFileName) THEN GOTO BufEnd  ' Nothing to do
      nErrCode = zFOpen(zFileName, 0, 2, hFile)
      IF nErrCode THEN
         nDone = nErrCode: nReading = 0
         EXIT FUNCTION
      END IF
      nRemaining = zFlof(hFile)                  ' Bytes count to be read
      nSizeFile = nRemaining                     ' Remember this
      IF nRemaining = 0 GOTO BufEnd              ' Nothing to read
      nBufSize = 16384
      sBuffer = SPACE$(nBufSize)
   END IF
   DO WHILE nRemaining                           ' While more in the file
      IF nCr = 0 THEN                            ' If no return was found
         IF nRemaining < nBufSize THEN           ' read only what remains.
            nBufSize = nRemaining                ' Resize the buffer.
            IF nBufSize < 1 THEN EXIT DO         ' Possible only if EOF 26.
            sBuffer = SPACE$(nBufSize)
         END IF
         nErrCode = zFGetAt(hFile, nLocation, sBuffer) ' Read a bloc.
         nLocation = nLocation + nBufSize        ' Here we are.
         nBufPos = 1                             ' Start at the beginning
      END IF                                     ' of that bloc.
      DO                                         ' Walk through buffer.
         nCr = INSTR(nBufPos, sBuffer, sCr)      ' Look for a Return.
         IF nLenCr = 0 THEN                      ' Do this to read files with
            nLenCr = 1                           ' or without LF after CR.
            IF ASC(sBuffer, nCr + 1) = 10 THEN nLenCr = 2
         END IF
         IF nCr THEN                             ' We found one.
            nSaveCr = nCr                        ' Save where
            FUNCTION = MID$(sBuffer, nBufPos, nCr - nBufPos)
            nBufPos = nCr + nLenCr               ' In case of LF skip it
            EXIT FUNCTION                        ' all done for now.
         ELSE                                    ' Back up in the file.
          ' If we reached the end of file and no nCr
          ' was found, return what remains in the string.
            IF nLocation >= nSizeFile THEN
             ' Assign function and trap ^Z
               FUNCTION = RTRIM$(MID$(sBuffer, nSaveCr + nLenCr), CHR$(26))
               nRemaining = nBufSize
               EXIT DO
            END IF
          ' 05-06-1998 () were missing on each side of (nLenCr - 1)
            nSlop = nBufSize - nSaveCr - (nLenCr - 1) ' Calc buffer excess.
            nRemaining = nRemaining + nSlop           ' Calc file excess.
            nLocation = nLocation - nSlop
         END IF
      LOOP WHILE nCr                                 ' While more in buffer.
      nRemaining = nRemaining - nBufSize
      nSaveCr = 0
   LOOP
END IF

BufEnd:
nReading = 0: nDone = -1: CALL zFClose(hFile)
END FUNCTION

FUNCTION WinMain (BYVAL hInstance     AS LONG, _
                  BYVAL hPrevInstance AS LONG, _
                  BYVAL lpCmdLine     AS ASCIIZ PTR, _
                  BYVAL iCmdShow      AS LONG) AS LONG

    LOCAL zFileName AS ASCIIZ * %MAX_PATH
    LOCAL nDone, lpBuffer, hOutput, nWritten AS LONG
    LOCAL sBuf AS STRING
    hOutput = GetStdHandle(%STD_OUTPUT_HANDLE)
    zFileName = "TEST.TXT"
    DO
        sBuf = zBufin(zFileName, nDone) + CHR$(13,10)
        lpBuffer = STRPTR(sBuf)
        CALL WriteConsole(hOutput, BYVAL lpBuffer, LEN(sBuf), nWritten, BYVAL 0&)
    LOOP UNTIL nDone

    CALL zBufin("", nDone) ' This can be used to close the handle
                           ' before full file completion
END FUNCTION
